﻿--[[
Name: Rating Buster
Revision: $Revision: 31849 $
Developed by: Whitetooth@Cenarius (hotdogee@bahamut.twbbs.org)
Description: Converts combat ratings in tooltips into normal percentages.
]]

--------------
-- Librarys --
--------------
local TipHooker = AceLibrary("TipHooker-1.0")
local StatLogic = AceLibrary("StatLogic-1.0")
local L = AceLibrary("AceLocale-2.2"):new("RatingBuster")


--------------------
-- AceAddon Setup --
--------------------
-- AceAddon Initialization
RatingBuster = AceLibrary("AceAddon-2.0"):new("AceDB-2.0", "AceConsole-2.0", "AceEvent-2.0", "AceDebug-2.0")
RatingBuster.title = "Rating Buster"
RatingBuster.version = "1.0.0 (r"..gsub("$Revision: 31849 $", "(%d+)", "%1")..")"
RatingBuster.date = gsub("$Date: 2007-04-04 08:31:24 -0400 (Wed, 04 Apr 2007) $", "^.-(%d%d%d%d%-%d%d%-%d%d).-$", "%1")


-------------------
-- Set Debugging --
-------------------
RatingBuster:SetDebugging(false)


-----------
-- Cache --
-----------
local cache = {}
setmetatable(cache, {__mode = "kv"}) -- weak table to enable garbage collection
local function clearCache()
	for k in pairs(cache) do
		cache[k] = nil
	end
end
--debug
--RatingBuster.cache = cache


---------------------
-- Local Variables --
---------------------
local _
local calcLevel, playerLevel
local profileDB, classDB -- Initialized in :OnInitialize()

-- Localize globals
local _G = getfenv(0)
local strfind = strfind
local strsub = strsub
local gsub = gsub
local pairs = pairs
local ipairs = ipairs
local type = type
local tinsert = tinsert
local tremove = tremove
local strsplit = strsplit
local strjoin = strjoin
local unpack = unpack
local tonumber = tonumber


---------------------
-- Saved Variables --
---------------------
-- Register DB
RatingBuster:RegisterDB("RatingBusterDB")
-- Default values
local profileDefault = {
	showItemLevel = true,
	showItemID = false,
	useRequiredLevel = false,
	customLevel = 0,
	textColor = {r = 1.0, g = 0.996,  b = 0.545, hex = "|cfffffe8b"},
	enableTextColor = true,
	enableStatMods = true,
}
--[[
Str -> AP, Block, Healing
Agi -> Crit, Dodge, AP, RAP, Armor
Sta -> Health, SpellDmg
Int -> Mana, SpellCrit, SpellDmg, Healing, MP5, RAP, Armor
Spi -> MP5, MP5NC, HP5, SpellDmg, Healing
--]]
local classDefault = {
	showAPFromStr = false,
	showBlockValueFromStr = false,
	showHealingFromStr = false, -- druid
	
	showCritFromAgi = true,
	showDodgeFromAgi = true,
	showAPFromAgi = false,
	showRAPFromAgi = false,
	showArmorFromAgi = false,
	
	showHealthFromSta = false,
	showSpellDmgFromSta = false, -- warlock
	
	showManaFromInt = false,
	showSpellCritFromInt = true,
	showSpellDmgFromInt = false, -- Druid, Mage, Paladin, Shaman, Warlock
	showHealingFromInt = false, -- Druid, Paladin, Shaman
	showMP5FromInt = false, -- Druid, Shaman
	showRAPFromInt = false, -- Hunter
	showArmorFromInt = false, -- Mage
	
	showMP5FromSpi = false, -- Druid, Mage, Priest
	showMP5NCFromSpi = false,
	showHP5FromSpi = false,
	showSpellDmgFromSpi = false, -- Priest
	showHealingFromSpi = false, -- Priest
}
local _, class = UnitClass("player")
if class == "DRUID" then
	classDefault.showHealingFromStr = true
end
if class == "PRIEST" or class == "MAGE" or class == "WARLOCK" then
	classDefault.showCritFromAgi = false
end
if class == "PRIEST" or class == "MAGE" or class == "WARLOCK" or class == "HUNTER" then
	classDefault.showDodgeFromAgi = false
end
if class == "WARLOCK" then
	classDefault.showSpellDmgFromSta = true
end
if class == "WARRIOR" or class == "ROGUE" or class == "HUNTER" then
	classDefault.showSpellCritFromInt = false
end
if class == "DRUID" or class == "MAGE" or class == "PALADIN" or class == "SHAMAN" or class == "WARLOCK" then
	classDefault.showSpellDmgFromInt = true
end
if class == "DRUID" or class == "PALADIN" or class == "SHAMAN" then
	classDefault.showHealingFromInt = true
end
if class == "DRUID" or class == "SHAMAN" then
	classDefault.showMP5FromInt = true
end
if class == "HUNTER" then
	classDefault.showRAPFromInt = true
end
if class == "MAGE" then
	classDefault.showArmorFromInt = true
end
if class == "DRUID" or class == "MAGE" or class == "PRIEST" then
	classDefault.showMP5FromSpi = true
end
if class == "PRIEST" then
	classDefault.showSpellDmgFromSpi = true
end
if class == "PRIEST" then
	classDefault.showHealingFromSpi = true
end

-- Register Defaults
RatingBuster:RegisterDefaults("profile", profileDefault)
RatingBuster:RegisterDefaults("class", classDefault)


---------------------------
-- Slash Command Options --
---------------------------
local consoleOptions = {
	type = 'group',
	args = {
		statmod = {
			type = 'toggle',
			name = L["Enable Stat Mods"],
			desc = L["Enable support for Stat Mods"],
			get = function() return profileDB.enableStatMods end,
			set = function(v)
				profileDB.enableStatMods = v
				-- clear cache
				clearCache()
			end,
		},
		itemid = {
			type = 'toggle',
			name = L["Show ItemID"],
			desc = L["Show the ItemID in tooltips"],
			get = function() return profileDB.showItemID end,
			set = function(v)
				profileDB.showItemID = v
				-- clear cache
				clearCache()
			end,
		},
		itemlevel = {
			type = 'toggle',
			name = L["Show ItemLevel"],
			desc = L["Show the ItemLevel in tooltips"],
			get = function() return profileDB.showItemLevel end,
			set = function(v)
				profileDB.showItemLevel = v
				-- clear cache
				clearCache()
			end,
		},
		usereqlv = {
			type = 'toggle',
			name = L["Use required level"],
			desc = L["Calculate using the required level if you are below the required level"],
			get = function() return profileDB.useRequiredLevel end,
			set = function(v) profileDB.useRequiredLevel = v end,
		},
		level = {
			type = 'range',
			name = L["Set level"],
			desc = L["Set the level used in calculations (0 = your level)"],
			get = function() return profileDB.customLevel end,
			set = function(v) profileDB.customLevel = v end,
			min = 0,
			max = 73, -- set to level cap + 3
			step = 1,
		},
		color = {
			type = 'group',
			name = L["Change text color"],
			desc = L["Changes the color of added text"],
			args = {
				pick = {
					type = 'execute',
					name = L["Pick color"],
					desc = L["Pick a color"],
					func = function()
						CloseMenus()
						ColorPickerFrame.func = function()
							profileDB.textColor.r, profileDB.textColor.g, profileDB.textColor.b = ColorPickerFrame:GetColorRGB();
							profileDB.textColor.hex = "|cff"..string.format("%02x%02x%02x", profileDB.textColor.r * 255, profileDB.textColor.g * 255, profileDB.textColor.b * 255)
							-- clear cache
							clearCache()
						end
						ColorPickerFrame:SetColorRGB(profileDB.textColor.r, profileDB.textColor.g, profileDB.textColor.b);
						ColorPickerFrame.previousValues = {r = profileDB.textColor.r, g = profileDB.textColor.g, b = profileDB.textColor.b};
						ColorPickerFrame:Show()
					end,
				},
				enable = {
					type = 'toggle',
					name = L["Enable color"],
					desc = L["Enable colored text"],
					get = function() return profileDB.enableTextColor end,
					set = function(v)
						profileDB.enableTextColor = v
						-- clear cache
						clearCache()
					end,
				},
			},
		},
		stat = {
			type = 'group',
			name = L["Stat"],
			desc = L["Changes the display of base stats"],
			args = {
				str = {
					type = 'group',
					name = L["Strength"],
					desc = L["Changes the display of Strength"],
					args = {
						ap = {
							type = 'toggle',
							name = L["Show Attack Power"],
							desc = L["Show Attack Power from Strength"],
							get = function() return classDB.showAPFromStr end,
							set = function(v)
								classDB.showAPFromStr = v
								-- clear cache
								clearCache()
							end,
						},
						block = {
							type = 'toggle',
							name = L["Show Block Value"],
							desc = L["Show Block Value from Strength"],
							get = function() return classDB.showBlockValueFromStr end,
							set = function(v)
								classDB.showBlockValueFromStr = v
								-- clear cache
								clearCache()
							end,
						},
						heal = {
							type = 'toggle',
							name = L["Show Healing"],
							desc = L["Show Healing from Strength"],
							get = function() return classDB.showHealingFromStr end,
							set = function(v)
								classDB.showHealingFromStr = v
								-- clear cache
								clearCache()
							end,
						},
					},
				},
				agi = {
					type = 'group',
					name = L["Agility"],
					desc = L["Changes the display of Agility"],
					args = {
						crit = {
							type = 'toggle',
							name = L["Show Crit"],
							desc = L["Show Crit chance from Agility"],
							get = function() return classDB.showCritFromAgi end,
							set = function(v)
								classDB.showCritFromAgi = v
								-- clear cache
								clearCache()
							end,
						},
						dodge = {
							type = 'toggle',
							name = L["Show Dodge"],
							desc = L["Show Dodge chance from Agility"],
							get = function() return classDB.showDodgeFromAgi end,
							set = function(v)
								classDB.showDodgeFromAgi = v
								-- clear cache
								clearCache()
							end,
						},
						ap = {
							type = 'toggle',
							name = L["Show Attack Power"],
							desc = L["Show Attack Power from Agility"],
							get = function() return classDB.showAPFromAgi end,
							set = function(v)
								classDB.showAPFromAgi = v
								-- clear cache
								clearCache()
							end,
						},
						rap = {
							type = 'toggle',
							name = L["Show Ranged Attack Power"],
							desc = L["Show Ranged Attack Power from Agility"],
							get = function() return classDB.showRAPFromAgi end,
							set = function(v)
								classDB.showRAPFromAgi = v
								-- clear cache
								clearCache()
							end,
						},
						armor = {
							type = 'toggle',
							name = L["Show Armor"],
							desc = L["Show Armor from Agility"],
							get = function() return classDB.showArmorFromAgi end,
							set = function(v)
								classDB.showArmorFromAgi = v
								-- clear cache
								clearCache()
							end,
						},
					},
				},
				sta = {
					type = 'group',
					name = L["Stamina"],
					desc = L["Changes the display of Stamina"],
					args = {
						hp = {
							type = 'toggle',
							name = L["Show Health"],
							desc = L["Show Health from Stamina"],
							get = function() return classDB.showHealthFromSta end,
							set = function(v)
								classDB.showHealthFromSta = v
								-- clear cache
								clearCache()
							end,
						},
						dmg = {
							type = 'toggle',
							name = L["Show Spell Damage"],
							desc = L["Show Spell Damage from Stamina"],
							get = function() return classDB.showSpellDmgFromSta end,
							set = function(v)
								classDB.showSpellDmgFromSta = v
								-- clear cache
								clearCache()
							end,
						},
					},
				},
				int = {
					type = 'group',
					name = L["Intellect"],
					desc = L["Changes the display of Intellect"],
					args = {
						spellcrit = {
							type = 'toggle',
							name = L["Show Spell Crit"],
							desc = L["Show Spell Crit chance from Intellect"],
							get = function() return classDB.showSpellCritFromInt end,
							set = function(v)
								classDB.showSpellCritFromInt = v
								-- clear cache
								clearCache()
							end,
						},
						mp = {
							type = 'toggle',
							name = L["Show Mana"],
							desc = L["Show Mana from Intellect"],
							get = function() return classDB.showManaFromInt end,
							set = function(v)
								classDB.showManaFromInt = v
								-- clear cache
								clearCache()
							end,
						},
						dmg = {
							type = 'toggle',
							name = L["Show Spell Damage"],
							desc = L["Show Spell Damage from Intellect"],
							get = function() return classDB.showSpellDmgFromInt end,
							set = function(v)
								classDB.showSpellDmgFromInt = v
								-- clear cache
								clearCache()
							end,
						},
						heal = {
							type = 'toggle',
							name = L["Show Healing"],
							desc = L["Show Healing from Intellect"],
							get = function() return classDB.showHealingFromInt end,
							set = function(v)
								classDB.showHealingFromInt = v
								-- clear cache
								clearCache()
							end,
						},
						mp5 = {
							type = 'toggle',
							name = L["Show Mana Regen"],
							desc = L["Show Mana Regen while casting from Intellect"],
							get = function() return classDB.showMP5FromInt end,
							set = function(v)
								classDB.showMP5FromInt = v
								-- clear cache
								clearCache()
							end,
						},
						rap = {
							type = 'toggle',
							name = L["Show Ranged Attack Power"],
							desc = L["Show Ranged Attack Power from Intellect"],
							get = function() return classDB.showRAPFromInt end,
							set = function(v)
								classDB.showRAPFromInt = v
								-- clear cache
								clearCache()
							end,
						},
						armor = {
							type = 'toggle',
							name = L["Show Armor"],
							desc = L["Show Armor from Intellect"],
							get = function() return classDB.showArmorFromInt end,
							set = function(v)
								classDB.showArmorFromInt = v
								-- clear cache
								clearCache()
							end,
						},
					},
				},
				spi = {
					type = 'group',
					name = L["Spirit"],
					desc = L["Changes the display of Spirit"],
					args = {
						mp5 = {
							type = 'toggle',
							name = L["Show Mana Regen"],
							desc = L["Show Mana Regen while casting from Spirit"],
							get = function() return classDB.showMP5FromSpi end,
							set = function(v)
								classDB.showMP5FromSpi = v
								-- clear cache
								clearCache()
							end,
						},
						mp5nc = {
							type = 'toggle',
							name = L["Show Mana Regen (not casting)"],
							desc = L["Show Mana Regen while NOT casting from Spirit"],
							get = function() return classDB.showMP5NCFromSpi end,
							set = function(v)
								classDB.showMP5NCFromSpi = v
								-- clear cache
								clearCache()
							end,
						},
						hp5 = {
							type = 'toggle',
							name = L["Show Health Regen"],
							desc = L["Show Health Regen from Spirit"],
							get = function() return classDB.showHP5FromSpi end,
							set = function(v)
								classDB.showHP5FromSpi = v
								-- clear cache
								clearCache()
							end,
						},
						dmg = {
							type = 'toggle',
							name = L["Show Spell Damage"],
							desc = L["Show Spell Damage from Spirit"],
							get = function() return classDB.showSpellDmgFromSpi end,
							set = function(v)
								classDB.showSpellDmgFromSpi = v
								-- clear cache
								clearCache()
							end,
						},
						heal = {
							type = 'toggle',
							name = L["Show Healing"],
							desc = L["Show Healing from Spirit"],
							get = function() return classDB.showHealingFromSpi end,
							set = function(v)
								classDB.showHealingFromSpi = v
								-- clear cache
								clearCache()
							end,
						},
					},
				},
			},
		},
	},
}

RatingBuster:RegisterChatCommand({"/rb", "/ratingbuster"}, consoleOptions)


-----------
-- Tools --
-----------
-- copyTable
local function copyTable(to, from)
	if to then
		for k in pairs(to) do
			to[k] = nil
		end
		setmetatable(to, nil)
	else
		to = {}
	end
	for k,v in pairs(from) do
		if type(k) == "table" then
			k = copyTable({}, k)
		end
		if type(v) == "table" then
			v = copyTable({}, v)
		end
		to[k] = v
	end
	setmetatable(to, getmetatable(from))
	return to
end


---------------------
-- Initializations --
---------------------
--[[ Loading Process Event Reference
{
ADDON_LOADED - When this addon is loaded
VARIABLES_LOADED - When all addons are loaded
PLAYER_LOGIN - Most information about the game world should now be available to the UI
}
--]]
-- OnInitialize(name) called at ADDON_LOADED
function RatingBuster:OnInitialize()
	profileDB = RatingBuster.db.profile
	classDB = RatingBuster.db.class
end

-- OnEnable() called at PLAYER_LOGIN
function RatingBuster:OnEnable()
	-- Hook item tooltips
	TipHooker:Hook(self.ProcessTooltip, "item")
	-- Initialize playerLevel
	playerLevel = UnitLevel("player")
	-- for setting a new level
	self:RegisterEvent("PLAYER_LEVEL_UP")
	-- Events that require cache clearing
	self:RegisterEvent("CHARACTER_POINTS_CHANGED", clearCache) -- talent point changed
	self:RegisterEvent("UNIT_AURA", 1) -- fire at most once every 1 second
end

function RatingBuster:OnDisable()
	-- Unhook item tooltips
	TipHooker:Unhook(self.ProcessTooltip, "item")
end

-- event = PLAYER_LEVEL_UP
-- arg1 = New player level
function RatingBuster:PLAYER_LEVEL_UP(newlevel)
	playerLevel = newlevel
	clearCache()
end

-- event = UNIT_AURA
-- arg1 = the UnitID of the entity
function RatingBuster:UNIT_AURA(unit)
	if (unit ~= "player") then return end
	clearCache()
end

--------------------------
-- Process Tooltip Core --
--------------------------
--[[
"+15 Agility"
-> "+15 Agility (+0.46% Crit)"
"+15 Crit Rating"
-> "+15 Crit Rating (+1.20%)"
"Equip: Increases your hit rating by 10."
-> "Equip: Increases your hit rating by 10 (+1.20%)."
--]]
function RatingBuster.ProcessTooltip(tooltip, name, link)
	-- Check if we're in standby mode
	--if not RatingBuster:IsActive() then return end
	---------------------------
	-- Set calculation level --
	---------------------------
	calcLevel = profileDB.customLevel
	if calcLevel == 0 then
		calcLevel = playerLevel
	end
	if profileDB.useRequiredLevel and link then
		local _, _, _, _, reqLevel = GetItemInfo(link)
		--RatingBuster:Print(link..", "..calcLevel)
		if reqLevel and calcLevel < reqLevel then
			calcLevel = reqLevel
		end
	end
	---------------------
	-- Tooltip Scanner --
	---------------------
	-- Loop through tooltip lines starting at line 2
	local tipTextLeft = tooltip:GetName().."TextLeft"
	for i = 2, tooltip:NumLines() do
		local fontString = _G[tipTextLeft..i]
		local text = fontString:GetText()
		if text then -- do nothing if we don't find a number
			-- Get data from cache if available
			local cacheID = text..calcLevel
			local cacheText = cache[cacheID]
			if cacheText then
				if cacheText ~= text then
					fontString:SetText(cacheText)
				end
			elseif strfind(text, "%d") then
				-- Initial pattern check, do nothing if not found
				-- Check for separators and bulid separatorTable
				local separatorTable = {}
				for _, sep in ipairs(L["separators"]) do
					if strfind(text, sep) then
						tinsert(separatorTable, sep)
					end
				end
				-- SplitDoJoin
				text = RatingBuster:SplitDoJoin(text, separatorTable)
				cache[cacheID] = text
				-- SetText
				fontString:SetText(text)
			end
		end
	end
	----------------------------
	-- Item Level and Item ID --
	----------------------------
	-- Check for ItemLevel addon, do nothing if found
	if not ItemLevel_AddInfo and (profileDB.showItemLevel or profileDB.showItemID) and link then
		if cache[link] then
			tooltip:AddLine(cache[link])
		else
			-- Get the Item ID from the link string
			local _, link, _, level = GetItemInfo(link)
			local _, _, id = strfind(link, "item:(%d+)")
			local newLine = ""
			if level and profileDB.showItemLevel then
				newLine = newLine..L["ItemLevel: "]..level
			end
			if id and profileDB.showItemID then
				if newLine ~= "" then
					newLine = newLine..", "
				end
				newLine = newLine..L["ItemID: "]..id
			end
			if newLine ~= "" then
				cache[link] = newLine
				tooltip:AddLine(newLine)
			end
		end
	end
	------------------
	-- Stat Summary --
	------------------
	--[[
	Health
	Mana
	Attack Power
	Ranged Attack Power
	Spell Damage
	Healing
	Hit Chance
	Crit Chance
	Spell Hit Chance
	Spell Crit Chance
	Dodge Chance
	Parry Chance
	Block Chance
	Armor
	Mana Regen(C/NC)
	Health Regen(C/OC)
	--]]
	---------------------
	-- Repaint tooltip --
	---------------------
	tooltip:Show()
end

---------------------------------------------------------------------------------
-- Recursive algorithm that divides a string into pieces using the separators in separatorTable,
-- processes them separately, then joins them back together
---------------------------------------------------------------------------------
-- text = "+24 Agility/+4 Stamina and +4 Spell Crit/+5 Spirit"
-- separatorTable = {"/", " and ", ","}
-- RatingBuster:SplitDoJoin("+24 Agility/+4 Stamina, +4 Dodge and +4 Spell Crit/+5 Spirit", {"/", " and ", ","})
--
function RatingBuster:SplitDoJoin(text, separatorTable)
	if type(separatorTable) == "table" and table.maxn(separatorTable) > 0 then
		local sep = tremove(separatorTable, 1)
		text =  gsub(text, sep, "@")
		text = {strsplit("@", text)}
		local processedText = {}
		local tempTable = {}
		for _, t in ipairs(text) do
			--self:Print(t[1])
			copyTable(tempTable, separatorTable)
			tinsert(processedText, self:SplitDoJoin(t, tempTable))
		end
		-- Join text
		return (gsub(strjoin("@", unpack(processedText)), "@", sep))
	else
		--self:Print(cacheID)
		return self:ProcessText(text)
	end
end

function RatingBuster:ProcessText(text)
	--self:Print(text)
	-- Check if test has a matching pattern
	for _, num in ipairs(L["numberPatterns"]) do
		-- Convert text to lower so we don't have to worry about same ratings with different cases
		local lowerText = string.lower(text)
		-- Capture the stat value
		local s, e, value = strfind(lowerText, num.pattern)
		if value then
			-- Capture the stat name
			for _, stat in ipairs(L["statList"]) do
				if strfind(lowerText, stat.pattern) then
					value = tonumber(value)
					local infoString = ""
					if type(stat.id) == "number" and stat.id >= 1 and stat.id <= 23 then
						--------------------
						-- Combat Ratings --
						--------------------
						-- Calculate stat value
						local effect = StatLogic:GetEffectFromRating(value, stat.id, calcLevel)
						--RatingBuster:Debug(reversedAmount..", "..amount..", "..v[2]..", "..RatingBuster.targetLevel)-- debug
						-- If rating is resilience, add a minus sign
						if stat.id >= 15 and stat.id <= 17 then
							effect = effect * -1
						end
						--RatingBuster.GetEffectFromRating(160, 18, 70)
						--self:Debug(text..", "..tostring(effect)..", "..value..", "..stat.id..", "..calcLevel)
						-- Build info string
						infoString = format("%+.2f", effect)
						if stat.id > 2 and stat.id < 21 then -- if rating is a percentage
							infoString = infoString.."%"
						end
					elseif stat.id == SPELL_STAT1_NAME then
						--------------
						-- Strength --
						--------------
						local statmod = 1
						if profileDB.enableStatMods then
							statmod = StatLogic:GetStatMod("MOD_STR")
							value = value * statmod
						end
						local infoTable = {}
						if classDB.showAPFromStr then
							local mod = StatLogic:GetStatMod("MOD_AP")
							local effect = value * StatLogic:GetAPPerStr(class) * mod
							if (mod ~= 1 or statmod ~= 1) and floor(effect * 10 + 0.5) > 0 then
								tinsert(infoTable, (gsub(L["$value AP"], "$value", format("%+.1f", effect))))
							elseif floor(effect + 0.5) > 0 then
								tinsert(infoTable, (gsub(L["$value AP"], "$value", format("%+.0f", effect))))
							end
						end
						if classDB.showBlockValueFromStr then
							local mod = StatLogic:GetStatMod("MOD_BLOCK_VALUE")
							local effect = value * StatLogic:GetBlockValuePerStr(class) * mod
							if floor(effect * 10 + 0.5) > 0 then
								tinsert(infoTable, (gsub(L["$value Block"], "$value", format("%+.1f", effect))))
							end
						end
						if classDB.showHealingFromStr then
							local mod = StatLogic:GetStatMod("MOD_HEALING")
							local effect = value * StatLogic:GetStatMod("ADD_HEALING_MOD_STR") * mod
							if floor(effect * 10 + 0.5) > 0 then
								tinsert(infoTable, (gsub(L["$value Heal"], "$value", format("%+.1f", effect))))
							end
						end
						infoString = strjoin(", ", unpack(infoTable))
					elseif stat.id == SPELL_STAT2_NAME then
						-------------
						-- Agility --
						-------------
						local statmod = 1
						if profileDB.enableStatMods then
							statmod = StatLogic:GetStatMod("MOD_AGI")
							value = value * statmod
						end
						local infoTable = {}
						if classDB.showAPFromAgi then
							local mod = StatLogic:GetStatMod("MOD_AP")
							local effect = value * StatLogic:GetAPPerAgi(class) * mod
							if (mod ~= 1 or statmod ~= 1) and floor(effect * 10 + 0.5) > 0 then
								tinsert(infoTable, (gsub(L["$value AP"], "$value", format("%+.1f", effect))))
							elseif floor(effect + 0.5) > 0 then
								tinsert(infoTable, (gsub(L["$value AP"], "$value", format("%+.0f", effect))))
							end
						end
						if classDB.showRAPFromAgi then
							local mod = StatLogic:GetStatMod("MOD_RANGED_AP")
							local effect = value * StatLogic:GetRAPPerAgi(class) * mod
							if (mod ~= 1 or statmod ~= 1) and floor(effect * 10 + 0.5) > 0 then
								tinsert(infoTable, (gsub(L["$value RAP"], "$value", format("%+.1f", effect))))
							elseif floor(effect + 0.5) > 0 then
								tinsert(infoTable, (gsub(L["$value RAP"], "$value", format("%+.0f", effect))))
							end
						end
						if classDB.showCritFromAgi then
							local effect = StatLogic:GetCritFromAgi(value, class, calcLevel)
							if effect > 0 then
								tinsert(infoTable, (gsub(L["$value% Crit"], "$value", format("%+.2f", effect))))
							end
						end
						if classDB.showDodgeFromAgi and (calcLevel == playerLevel) then
							local effect = StatLogic:GetDodgeFromAgi(value)
							if effect > 0 then
								tinsert(infoTable, (gsub(L["$value% Dodge"], "$value", format("%+.2f", effect))))
							end
						end
						if classDB.showArmorFromAgi then
							local effect = value * 2
							if effect > 0 then
								tinsert(infoTable, (gsub(L["$value Armor"], "$value", format("%+.0f", effect))))
							end
						end
						infoString = strjoin(", ", unpack(infoTable))
					elseif stat.id == SPELL_STAT3_NAME then
						-------------
						-- Stamina --
						-------------
						local statmod = 1
						if profileDB.enableStatMods then
							statmod = StatLogic:GetStatMod("MOD_STA")
							value = value * statmod
						end
						local infoTable = {}
						if classDB.showHealthFromSta then
							local mod = StatLogic:GetStatMod("MOD_HEALTH")
							local effect = value * 10 * mod -- 10 Health per Sta
							if (mod ~= 1 or statmod ~= 1) and floor(effect * 10 + 0.5) > 0 then
								tinsert(infoTable, (gsub(L["$value HP"], "$value", format("%+.1f", effect))))
							elseif floor(effect + 0.5) > 0 then
								tinsert(infoTable, (gsub(L["$value HP"], "$value", format("%+.0f", effect))))
							end
						end
						if classDB.showSpellDmgFromSta then
							local mod = StatLogic:GetStatMod("MOD_SPELL_DMG")
							local effect = value * StatLogic:GetStatMod("ADD_SPELL_DMG_MOD_STA") * mod
							if floor(effect * 10 + 0.5) > 0 then
								tinsert(infoTable, (gsub(L["$value Dmg"], "$value", format("%+.1f", effect))))
							end
						end
						infoString = strjoin(", ", unpack(infoTable))
					elseif stat.id == SPELL_STAT4_NAME then
						---------------
						-- Intellect --
						---------------
						local statmod = 1
						if profileDB.enableStatMods then
							statmod = StatLogic:GetStatMod("MOD_INT")
							value = value * statmod
						end
						local infoTable = {}
						if classDB.showManaFromInt then
							local mod = StatLogic:GetStatMod("MOD_MANA")
							local effect = value * 15 * mod -- 15 Mana per Int
							if (mod ~= 1 or statmod ~= 1) and floor(effect * 10 + 0.5) > 0 then
								tinsert(infoTable, (gsub(L["$value MP"], "$value", format("%+.1f", effect))))
							elseif floor(effect + 0.5) > 0 then
								tinsert(infoTable, (gsub(L["$value MP"], "$value", format("%+.0f", effect))))
							end
						end
						if classDB.showSpellCritFromInt then
							local effect = StatLogic:GetSpellCritFromInt(value, class, calcLevel)
							if effect > 0 then
								tinsert(infoTable, (gsub(L["$value% Spell Crit"], "$value", format("%+.2f", effect))))
							end
						end
						if classDB.showSpellDmgFromInt then
							local mod = StatLogic:GetStatMod("MOD_SPELL_DMG")
							local effect = value * StatLogic:GetStatMod("ADD_SPELL_DMG_MOD_INT") * mod
							if floor(effect * 10 + 0.5) > 0 then
								tinsert(infoTable, (gsub(L["$value Dmg"], "$value", format("%+.1f", effect))))
							end
						end
						if classDB.showHealingFromInt then
							local mod = StatLogic:GetStatMod("MOD_HEALING")
							local effect = value * StatLogic:GetStatMod("ADD_HEALING_MOD_INT") * mod
							if floor(effect * 10 + 0.5) > 0 then
								tinsert(infoTable, (gsub(L["$value Heal"], "$value", format("%+.1f", effect))))
							end
						end
						if classDB.showMP5FromInt then
							local effect = value * StatLogic:GetStatMod("ADD_MANA_REG_MOD_INT")
							if floor(effect * 10 + 0.5) > 0 then
								tinsert(infoTable, (gsub(L["$value MP5"], "$value", format("%+.1f", effect))))
							end
						end
						if classDB.showRAPFromInt then
							local mod = StatLogic:GetStatMod("MOD_RANGED_AP")
							local effect = value * StatLogic:GetStatMod("ADD_RANGED_AP_MOD_INT") * mod
							if floor(effect * 10 + 0.5) > 0 then
								tinsert(infoTable, (gsub(L["$value RAP"], "$value", format("%+.1f", effect))))
							end
						end
						if classDB.showArmorFromInt then
							local effect = value * StatLogic:GetStatMod("ADD_ARMOR_MOD_INT")
							if floor(effect + 0.5) > 0 then
								tinsert(infoTable, (gsub(L["$value Armor"], "$value", format("%+.0f", effect))))
							end
						end
						infoString = strjoin(", ", unpack(infoTable))
					elseif stat.id == SPELL_STAT5_NAME then
						------------
						-- Spirit --
						------------
						local statmod = 1
						if profileDB.enableStatMods then
							statmod = StatLogic:GetStatMod("MOD_SPI")
							value = value * statmod
						end
						local infoTable = {}
						if classDB.showMP5FromSpi then
							local mod = StatLogic:GetStatMod("ADD_MANA_REG_MOD_NORMAL_MANA_REG")
							local effect = StatLogic:GetNormalManaRegenFromSpi(value, class) * mod
							if floor(effect * 10 + 0.5) > 0 then
								tinsert(infoTable, (gsub(L["$value MP5"], "$value", format("%+.1f", effect))))
							end
						end
						if classDB.showMP5NCFromSpi then
							local effect = StatLogic:GetNormalManaRegenFromSpi(value, class)
							if floor(effect * 10 + 0.5) > 0 then
								tinsert(infoTable, (gsub(L["$value MP5(NC)"], "$value", format("%+.1f", effect))))
							end
						end
						if classDB.showHP5FromSpi then
							local effect = StatLogic:GetHealthRegenFromSpi(value, class)
							if floor(effect * 10 + 0.5) > 0 then
								tinsert(infoTable, (gsub(L["$value HP5"], "$value", format("%+.1f", effect))))
							end
						end
						if classDB.showSpellDmgFromSpi then
							local mod = StatLogic:GetStatMod("MOD_SPELL_DMG")
							local effect = value * StatLogic:GetStatMod("ADD_SPELL_DMG_MOD_SPI") * mod
							if floor(effect * 10 + 0.5) > 0 then
								tinsert(infoTable, (gsub(L["$value Dmg"], "$value", format("%+.1f", effect))))
							end
						end
						if classDB.showHealingFromSpi then
							local mod = StatLogic:GetStatMod("MOD_HEALING")
							local effect = value * StatLogic:GetStatMod("ADD_HEALING_MOD_SPI") * mod
							if floor(effect * 10 + 0.5) > 0 then
								tinsert(infoTable, (gsub(L["$value Heal"], "$value", format("%+.1f", effect))))
							end
						end
						infoString = strjoin(", ", unpack(infoTable))
					end
					if infoString ~= "" then
						-- Add parenthesis
						infoString = "("..infoString..")"
						-- Add Color
						if profileDB.enableTextColor then
							infoString = profileDB.textColor.hex..infoString.."|r"
						end
						-- Build replacement string
						if num.addInfo == "AfterNumber" then -- Add after number
							infoString = gsub(infoString, "%%", "%%%%%%%%")
							infoString = gsub(strsub(text, s, e), "%d+", "%0 "..infoString)
						else -- Add after stat
							infoString = gsub(infoString, "%%", "%%%%")
							s, e = strfind(lowerText, stat.pattern)
							infoString = "%0 "..infoString
						end
						-- Insert info into text
						return (gsub(text, strsub(text, s, e), infoString, 1)) -- because gsub has 2 return values, but we only want 1
					end
					return text
				end
			end
		end
	end
	return text
end

-- RatingBuster:Bench(1000)
---------
-- self:SplitDoJoin("+24 Agility/+4 Stamina, +4 Dodge and +4 Spell Crit/+5 Spirit", {"/", " and ", ","})
-- 1000 times: 0.16 - 0.18 without Compost
-- 1000 times: 0.22 - 0.24 with Compost
---------
-- RatingBuster.ProcessTooltip(ItemRefTooltip, link)
-- 1000 times: 0.31 sec - 0.7.6
-- 1000 times: 0.29 sec - 0.
-- 1000 times: 0.24 sec - 0.8.58.0
---------
-- strjoin 1000000 times: 0.46
-- ..      1000000 times: 0.27
--------------
function RatingBuster:Bench(k)
	local t1 = GetTime()
	local link = GetInventoryItemLink("player", 12)
	for i = 1, k, 1 do
		---------------------------------------------------------------------------
		--self:SplitDoJoin("+24 Agility/+4 Stamina, +4 Dodge and +4 Spell Crit/+5 Spirit", {"/", " and ", ","})
		---------------------------------------------------------------------------
		ItemRefTooltip:SetInventoryItem("player", 12)
		RatingBuster.ProcessTooltip(ItemRefTooltip, link)
		---------------------------------------------------------------------------
		--ItemRefTooltip:SetScript("OnTooltipSetItem", function(frame, ...) RatingBuster:Print("OnTooltipSetItem") end)
		----------------------------------------------------------------------
		--local h = strjoin("", "test", "123")
		--local h = "test".."123"
		--------------------------------------------------------------------------------
	end
	return GetTime() - t1
end
